/** ANSI C ANTLR v3 grammar

Translated from Jutta Degener's 1995 ANSI C yacc grammar by Terence Parr
July 2006.  The lexical rules were taken from the Java grammar.

Jutta says: "In 1985, Jeff Lee published his Yacc grammar (which
is accompanied by a matching Lex specification) for the April 30, 1985 draft
version of the ANSI C standard.  Tom Stockfisch reposted it to net.sources in
1987; that original, as mentioned in the answer to question 17.25 of the
comp.lang.c FAQ, can be ftp'ed from ftp.uu.net,
   file usenet/net.sources/ansi.c.grammar.Z.
I intend to keep this version as close to the current C Standard grammar as
possible; please let me know if you discover discrepancies. Jutta Degener, 1995"

Generally speaking, you need symbol table info to parse C; typedefs
define types and then IDENTIFIERS are either types or plain IDs.  I'm doing
the min necessary here tracking only type names.  This is a good example
of the global scope (called Symbols).  Every rule that declares its usage
of Symbols pushes a new copy on the stack effectively creating a new
symbol scope.  Also note rule declaration declares a rule scope that
lets any invoked rule see isTypedef boolean.  It's much easier than
passing that info down as parameters.  Very clean.  Rule
direct_declarator can then easily determine whether the IDENTIFIER
should be declared as a type name.

I have only tested this on a single file, though it is 3500 lines.

This grammar requires ANTLR v3.0.1 or higher.

Terence Parr
July 2006
*/

grammar C;

options {
	language = C;
	output = AST;
	backtrack = true;
	memoize = true;
	k = 2;
}

tokens {
	AST_TRANSLATION_UNIT;
	AST_FUNCTION_DECLARATION;
	AST_FUNCTION_DECLARATION_HEAD;
	AST_VARIABLE_DECLARATION_OR_TYPE;
	AST_VARIABLE_DECLARATION_OR_TYPE_HEAD;
	AST_TYPE;
	AST_BITFIELD;
	AST_STORAGE_CLASS_SPECIFIER;
	AST_BASIC_TYPE;
	AST_TYPEDEF_TYPE;
	AST_ENUM_FIELD;
	AST_ARRAY_TYPE;
	AST_POINTER_TYPE;
	AST_PARAMETER_LIST;
	AST_CAST_EXPRESSION;
	AST_UNARY_EXPRESSION;
	AST_POSTFIX_EXPRESSION;
	AST_FUNCTION_CALL;
	AST_BUILTIN_VA_ARG_CALL;
	AST_ALIGNOF_CALL;
	AST_PARENTHESIZED_EXPRESSION;
	AST_VARIABLE_EXPRESSION;
	AST_ARRAY_INITIALIZER;
	AST_HEX_CONSTANT;
	AST_OCTAL_CONSTANT;
	AST_DECIMAL_CONSTANT;
	AST_CHARACTER_CONSTANT;
	AST_STRING_CONSTANT;
	AST_FLOATING_POINT_CONSTANT;
	AST_EXPRESSION;
	AST_ASSIGNMENT_EXPRESSION;
	AST_CONDITIONAL_EXPRESSION;
	AST_STATEMENT;
	AST_LABELED_STATEMENT;
	AST_COMPOUND_STATEMENT;
	AST_EXPRESSION_STATEMENT;
	AST_IF_STATEMENT;
	AST_WHILE_STATEMENT;
	AST_JUMP_STATEMENT;
	AST_ASSEMBLER_STATEMENT;
	AST_ASM_TEMPLATE;
	AST_ASM_OPERAND_LIST;
	AST_ASM_CLOBBERED;
	AST_TYPE_QUALIFIER;
	AST_GCC_ATTRIBUTE;

	AST_BT_VOID;
	AST_BT_BOOL;
	AST_BT_CHAR;
	AST_BT_SHORT;
	AST_BT_INT;
	AST_BT_LONG;
	AST_BT_LONG_LONG;
	AST_BT_UNSIGNED_CHAR;
	AST_BT_UNSIGNED_SHORT;
	AST_BT_UNSIGNED_INT;
	AST_BT_UNSIGNED_LONG;
	AST_BT_UNSIGNED_LONG_LONG;
	AST_BT_FLOAT;
	AST_BT_DOUBLE;
	AST_BT_LONG_DOUBLE;
	AST_BT_BUILTIN_VA_LIST;
}

@lexer::includes {
	#include "error_handlers.h"
}

@lexer::apifuncs {
	RECOGNIZER->displayRecognitionError = &eh_lexer_displayRecognitionError;
}

@parser::includes {
	#include "error_handlers.h"
}

@parser::apifuncs {
	RECOGNIZER->displayRecognitionError = &eh_parser_displayRecognitionError;
}

@members {
	ANTLR3_BOOLEAN typedef_active;
	pANTLR3_HASH_TABLE typedef_names;
	pANTLR3_COMMON_TOKEN token_bak;

	void setup_typedef_set()
	{
		typedef_active = ANTLR3_FALSE;
		typedef_names = antlr3HashTableNew(100);
	}

	void dispose_typedef_set()
	{
		typedef_names->free(typedef_names);
	}

	void add_typedef(pANTLR3_COMMON_TOKEN t)
	{
		if(typedef_active)
		{
			typedef_names->put(typedef_names, t->getText(t)->chars, (void*)1, NULL);
			typedef_active = ANTLR3_FALSE;
		}
	}

	ANTLR3_BOOLEAN is_typedef(pANTLR3_COMMON_TOKEN t)
	{
		 return typedef_names->get(typedef_names, t->getText(t)->chars) != NULL;
	}

	void set_source_file(pANTLR3_BASE_TREE external_declaration_node, pANTLR3_COMMON_TOKEN first_real_token)
	{
		// Go through the raw character stream. The "#line ..." info might not contain the source file,
		// then we have to look for it going backwards. But since we parse the C file in external
		// declaration chunks, we don't have the tokens anymore. Only the character stream.

		char *c_source = (char *)first_real_token->start;
		char *s;
		while(1)
		{
			if(strncmp(--c_source, "#line ", 6) != 0) continue;
			s = c_source + 6;

			// skip the number block
			if(!(*s >= '0' && *s <= '9')) continue;
			while(*s >= '0' && *s <= '9') s++;
			s++;

			// set the source file
			if(*s != '"') continue;
			s++;
			external_declaration_node->u = s;
			return;
		}
	}
}

translation_unit : translation_unit2 -> ^(AST_TRANSLATION_UNIT translation_unit2);
translation_unit2
	:
		(
			{token_bak = LT(1);}
			external_declaration
			{set_source_file($external_declaration.tree, token_bak);}
		)* EOF!
	;

/** Either a function definition or any other kind of C decl/def.
 *  The LL(*) analysis algorithm fails to deal with this due to
 *  recursion in the declarator rules.  I'm putting in a
 *  manual predicate here so that we don't backtrack over
 *  the entire function.  Further, you get a better error
 *  as errors within the function itself don't make it fail
 *  to predict that it's a function.  Weird errors previously.
 *  Remember: the goal is to avoid backtrack like the plague
 *  because it makes debugging, actions, and errors harder.
 *
 *  Note that k=1 results in a much smaller predictor for the
 *  fixed lookahead; k=2 made a few extra thousand lines. ;)
 *  I'll have to optimize that in the future.
 */
external_declaration
		options {k=1;}
	: typedef_declaration
	| struct_declaration
	| enum_declaration
	| function_declaration
	| variable_declaration_or_type ';'!
	;

typedef_declaration
		@init { typedef_active = ANTLR3_TRUE; }
	: 'typedef'^ (variable_declaration_or_type ';'! | function_declaration)
	;

struct_declaration
	: ('struct'|'union')^ IDENTIFIER ('{' ((bitfield | variable_declaration_or_type) ';'!)* '}'! gcc_attribute_specifier?)? ';'!
	;

enum_declaration
	: 'enum'^ IDENTIFIER ('{'! enum_field (','! enum_field)* '}'!)? ';'!
	;

variable_declaration_or_type : variable_declaration_or_type2 -> ^(AST_VARIABLE_DECLARATION_OR_TYPE variable_declaration_or_type2); // without an identifier it's a type
variable_declaration_or_type2
	: storage_class_specifier* type gcc_attribute_specifier?
		(
			variable_declaration_or_type_head gcc_attribute_specifier? ('='! (assignment_expression | array_initializer))?
		|
			'('! pointer_type (IDENTIFIER {add_typedef($IDENTIFIER);})? array_type? ')'! array_type // POINTER [ARRAY [type]]
		)?
	;

variable_declaration_or_type_head : variable_declaration_or_type_head2 -> ^(AST_VARIABLE_DECLARATION_OR_TYPE_HEAD variable_declaration_or_type_head2);
variable_declaration_or_type_head2
	: IDENTIFIER {add_typedef($IDENTIFIER);} array_type?
	| array_type
	| '('!* gcc_attribute_specifier? '('!* pointer_type variable_declaration_or_type_head? ')'!+ '('! parameter_list? ')'!
	;

function_declaration : function_declaration2 -> ^(AST_FUNCTION_DECLARATION function_declaration2);
function_declaration2
	: storage_class_specifier* gcc_attribute_specifier? type gcc_attribute_specifier? function_declaration_head ('{'! (variable_declaration_or_type ';'!)* compound_statement '}'! | gcc_attribute_specifier? ';'!)
	;

function_declaration_head : function_declaration_head2 -> ^(AST_FUNCTION_DECLARATION_HEAD function_declaration_head2);
function_declaration_head2
	: '('!? (
			gcc_attribute_specifier? '__cdecl'!? IDENTIFIER {add_typedef($IDENTIFIER);} |
			pointer_type function_declaration_head
		) ')'!? '('! parameter_list? ')'!
	;

bitfield : bitfield2 -> ^(AST_BITFIELD bitfield2);
bitfield2
	: (basic_type | typedef_type | enum_type) type_qualifier!? IDENTIFIER? ':'! decimal_constant gcc_attribute_specifier?
	;

storage_class_specifier : storage_class_specifier2 -> ^(AST_STORAGE_CLASS_SPECIFIER storage_class_specifier2);
storage_class_specifier2
	: 'extern'
	| 'static'
	| 'auto'
	| 'register'
	| '__inline'
	;

type : type2 -> ^(AST_TYPE type2);
type2
	: (basic_type | struct_type | enum_type | typedef_type) type_qualifier? pointer_type?
	;

type_qualifier : type_qualifier2 -> ^(AST_TYPE_QUALIFIER type_qualifier2);
type_qualifier2
	: 'const'
	| 'volatile'
	| 'const' 'volatile'
	;

basic_type : basic_type2 -> ^(AST_BASIC_TYPE basic_type2);
basic_type2
	: 'void'              -> ^(AST_BT_VOID)
	| '_Bool'             -> ^(AST_BT_BOOL)
	| '__builtin_va_list' -> ^(AST_BT_BUILTIN_VA_LIST)

	| 'float'         -> ^(AST_BT_FLOAT)       // REAL_32
	| 'double'        -> ^(AST_BT_DOUBLE)      // REAL_64
	| 'long' 'double' -> ^(AST_BT_LONG_DOUBLE) // ?

	| 'signed'? 'char' -> ^(AST_BT_CHAR)      // INTEGER_8
	| 'short'          -> ^(AST_BT_SHORT)     // INTEGER_16
	| 'int'            -> ^(AST_BT_INT)       // INTEGER_32
	| 'long' 'long'    -> ^(AST_BT_LONG_LONG) // INTEGER_64
	| 'long'           -> ^(AST_BT_LONG)      // INTEGER_32

	| 'unsigned' 'char'        -> ^(AST_BT_UNSIGNED_CHAR)      // NATURAL_8
	| 'unsigned' 'short'       -> ^(AST_BT_UNSIGNED_SHORT)     // NATURAL_16
	| 'unsigned' 'int'         -> ^(AST_BT_UNSIGNED_INT)       // NATURAL_32
	| 'unsigned' 'long' 'long' -> ^(AST_BT_UNSIGNED_LONG_LONG) // NATURAL_64
	| 'unsigned' 'long'        -> ^(AST_BT_UNSIGNED_LONG)      // NATURAL_32
	;

struct_type
	: ('struct'|'union')^ IDENTIFIER
	;

enum_type
	: 'enum'^ IDENTIFIER
	;

typedef_type : typedef_type2 -> ^(AST_TYPEDEF_TYPE typedef_type2);
typedef_type2
    : {is_typedef(LT(1))}? IDENTIFIER
    ;

enum_field : enum_field2 -> ^(AST_ENUM_FIELD enum_field2);
enum_field2
	: IDENTIFIER '=' expression
	;

array_type : array_type2 -> ^(AST_ARRAY_TYPE array_type2);
array_type2
	: ('[' expression? ']')+
	;

pointer_type : pointer_type2 -> ^(AST_POINTER_TYPE pointer_type2);
pointer_type2
	: '*'! type_qualifier pointer_type?
	| '*'! pointer_type
	| '*'!
	;

parameter_list : parameter_list2 -> ^(AST_PARAMETER_LIST parameter_list2);
parameter_list2
	: variable_declaration_or_type (','! variable_declaration_or_type)* (','! '...')?
	;

array_initializer : array_initializer2 -> ^(AST_ARRAY_INITIALIZER array_initializer2);
array_initializer2
	: '{'! ((array_initializer|assignment_expression) (','! (array_initializer|assignment_expression))*)? '}'!
	;

// E x p r e s s i o n s

hex_constant : hex_constant2 -> ^(AST_HEX_CONSTANT hex_constant2);
hex_constant2
	: HEX_LITERAL
	;

octal_constant : octal_constant2 -> ^(AST_OCTAL_CONSTANT octal_constant2);
octal_constant2
	: OCTAL_LITERAL
	;

decimal_constant : decimal_constant2 -> ^(AST_DECIMAL_CONSTANT decimal_constant2);
decimal_constant2
	: DECIMAL_LITERAL
	;

character_constant : character_constant2 -> ^(AST_CHARACTER_CONSTANT character_constant2);
character_constant2
	: CHARACTER_LITERAL
	;

string_constant : string_constant2 -> ^(AST_STRING_CONSTANT string_constant2);
string_constant2
	: STRING_LITERAL
	;

floating_point_constant : floating_point_constant2 -> ^(AST_FLOATING_POINT_CONSTANT floating_point_constant2);
floating_point_constant2
	: FLOATING_POINT_LITERAL
	;

expression : expression2 -> ^(AST_EXPRESSION expression2);
expression2
	: assignment_expression
	;

assignment_expression
	: assignment_expression2 -> ^(AST_ASSIGNMENT_EXPRESSION assignment_expression2)
	| conditional_expression
	;
assignment_expression2
	: unary_expression ('='|'*='|'/='|'%='|'+='|'-='|'<<='|'>>='|'&='|'^='|'|=') assignment_expression
	;

conditional_expression
	: conditional_expression2 -> ^(AST_CONDITIONAL_EXPRESSION conditional_expression2)
	| logical_expression
	;
conditional_expression2
	: logical_expression '?' expression ':' conditional_expression
	;

logical_expression
	: inclusive_or_expression (('&&'|'||')^ inclusive_or_expression)*
	;

inclusive_or_expression
	: exclusive_or_expression ('|'^ exclusive_or_expression)*
	;

exclusive_or_expression
	: and_expression ('^'^ and_expression)*
	;

and_expression
	: equality_expression ('&'^ equality_expression)*
	;

equality_expression
	: relational_expression (('=='|'!=')^ relational_expression)*
	;

relational_expression
	: shift_expression (('<'|'>'|'<='|'>=')^ shift_expression)*
	;

shift_expression
	: additive_expression (('<<'|'>>')^ additive_expression)*
	;

additive_expression
	: multiplicative_expression (('+'|'-')^ multiplicative_expression)*
	;

multiplicative_expression
	: cast_expression (('*'|'/'|'%')^ cast_expression)*
	;

cast_expression
	: cast_expression2 -> ^(AST_CAST_EXPRESSION cast_expression2)
	| unary_expression
	;
cast_expression2
	: '('! variable_declaration_or_type ')'! cast_expression
	;

unary_expression
	: unary_expression2 -> ^(AST_UNARY_EXPRESSION unary_expression2)
	| sizeof_expression
	| builtin_va_arg_call // Needs to be before "function_call" to correctly process calls like "__builtin_va_arg(foo, int64_t)".
	| alignof_call // see comment above
	| function_call
	| postfix_expression
	;
unary_expression2
	: '++' unary_expression
	| '--' unary_expression
	| ('&'|'*'|'+'|'-'|'~'|'!') cast_expression
	;

sizeof_expression
	: 'sizeof'^ '('! (variable_declaration_or_type | expression) ')'!
	;

function_call : function_call2 -> ^(AST_FUNCTION_CALL function_call2);
function_call2
	:	(IDENTIFIER | '('! expression ')'!) '('! (assignment_expression (','! assignment_expression)*)? ')'!
	;

builtin_va_arg_call : builtin_va_arg_call2 -> ^(AST_BUILTIN_VA_ARG_CALL builtin_va_arg_call2);
builtin_va_arg_call2
	: {strncmp("__builtin_va_arg", (char *)LT(1)->getStartIndex(LT(1)), 16) == 0}? IDENTIFIER! '('! expression ','! variable_declaration_or_type ')'!
	;

alignof_call : alignof_call2 -> ^(AST_ALIGNOF_CALL alignof_call2);
alignof_call2
	: {strncmp("__alignof__", (char *)LT(1)->getStartIndex(LT(1)), 11) == 0}? IDENTIFIER! '('! variable_declaration_or_type ')'!
	;

postfix_expression
	: postfix_expression2 -> ^(AST_POSTFIX_EXPRESSION postfix_expression2)
	| primary_expression
	;
postfix_expression2
	: primary_expression ('++' | '--')
	| primary_expression ('[' expression ']' | '.' IDENTIFIER | '->' IDENTIFIER)+
	;

primary_expression
	: '(' expression ')' -> ^(AST_PARENTHESIZED_EXPRESSION expression)
	| IDENTIFIER -> ^(AST_VARIABLE_EXPRESSION IDENTIFIER)
	| hex_constant
	| octal_constant
	| decimal_constant
	| character_constant
	| string_constant
	| floating_point_constant
	;

// S t a t e m e n t s

statement : statement2 -> ^(AST_STATEMENT statement2);
statement2
	: labeled_statement
	| compound_statement
	| expression_statement
	| if_statement
	| switch_statement
	| while_statement
	| jump_statement
	| switch_case
	| switch_default
	| assembler_statement
	;

labeled_statement : labeled_statement2 -> ^(AST_LABELED_STATEMENT labeled_statement2);
labeled_statement2
	: IDENTIFIER ':'! statement
	;

compound_statement : compound_statement2 -> ^(AST_COMPOUND_STATEMENT compound_statement2);
compound_statement2
	: '{' statement* '}'
	;

expression_statement : expression_statement2 -> ^(AST_EXPRESSION_STATEMENT expression_statement2);
expression_statement2
	: ';'
	| expression ';'
	;

if_statement : if_statement2 -> ^(AST_IF_STATEMENT if_statement2);
if_statement2
	: 'if'! '('! expression ')'! statement (options {k=1; backtrack=false;}:'else'! statement)?
	;

switch_statement
	: 'switch'^ '('! expression ')'! compound_statement
	;

switch_case
	: 'case'^ expression ':'
	;

switch_default
	: 'default'^ ':'
	;

while_statement : while_statement2 -> ^(AST_WHILE_STATEMENT while_statement2);
while_statement2
	: 'while' '(' expression ')' compound_statement
	;

jump_statement : jump_statement2 -> ^(AST_JUMP_STATEMENT jump_statement2);
jump_statement2
	: 'goto' IDENTIFIER ';'
	| 'continue' ';'
	| 'break' ';'
	| 'return' expression? ';'
	;

assembler_statement : assembler_statement2 -> ^(AST_ASSEMBLER_STATEMENT assembler_statement2);
assembler_statement2
	: '__asm__ '! 'volatile'!? '('! asm_template (':' asm_operand_list? (':' asm_operand_list? (':' asm_clobbered?)?)?)? ')'! ';'!
	;

asm_template : asm_template2 -> ^(AST_ASM_TEMPLATE asm_template2);
asm_template2
	: STRING_LITERAL+
	;

asm_operand_list : asm_operand_list2 -> ^(AST_ASM_OPERAND_LIST asm_operand_list2);
asm_operand_list2
	: STRING_LITERAL '('! expression ')'! (','! STRING_LITERAL  '('! expression ')'!)*
	;

asm_clobbered : asm_clobbered2 -> ^(AST_ASM_CLOBBERED asm_clobbered2);
asm_clobbered2
	: STRING_LITERAL (','! STRING_LITERAL)*
	;

gcc_attribute_specifier
	: '__attribute__'^ '('! '('! gcc_attribute (','! gcc_attribute)* ')'! ')'!
	;

gcc_attribute : gcc_attribute2 -> ^(AST_GCC_ATTRIBUTE gcc_attribute2);
gcc_attribute2
	: IDENTIFIER ('('! expression (','! expression)* ')'!)?
	;


RESTRICT_INFO
	: '__restrict' {SKIP();}
	;

ASM
	: '__asm__(' .+ ') ' {SKIP();}
	;


IDENTIFIER
	:	LETTER (LETTER|'0'..'9')*
	;

fragment
LETTER
	:	'$'
	|	'A'..'Z'
	|	'a'..'z'
	|	'_'
	;

CHARACTER_LITERAL
	: '\'' ( EscapeSequence | ~('\''|'\\') ) '\''
	;

STRING_LITERAL
	:  'L'? '"' ( EscapeSequence | ~('\\'|'"') )* '"'
	;

HEX_LITERAL
	: '0' ('x'|'X') HexDigit+ IntegerTypeSuffix?
	| '0' ('x'|'X') HexDigit+ '.' HexDigit* ('p'|'P') ('+'|'-')? ('0'..'9')+ FloatTypeSuffix?
	| '0' ('x'|'X')           '.' HexDigit+ ('p'|'P') ('+'|'-')? ('0'..'9')+ FloatTypeSuffix?
	| '0' ('x'|'X') HexDigit+               ('p'|'P') ('+'|'-')? ('0'..'9')+ FloatTypeSuffix?
	;

DECIMAL_LITERAL : ('0' | '1'..'9' '0'..'9'*) IntegerTypeSuffix? ;

OCTAL_LITERAL : '0' ('0'..'7')+ IntegerTypeSuffix? ;

fragment
HexDigit : ('0'..'9'|'a'..'f'|'A'..'F') ;

fragment
IntegerTypeSuffix
	: ('u'|'U')
	| ('l'|'L'|'ll'|'LL')
	| ('u'|'U') ('l'|'L'|'ll'|'LL')
	;

FLOATING_POINT_LITERAL
	: ('0'..'9')+ '.' ('0'..'9')* Exponent? FloatTypeSuffix?
	| '.' ('0'..'9')+ Exponent? FloatTypeSuffix?
	| ('0'..'9')+ ( Exponent FloatTypeSuffix? | FloatTypeSuffix)
	;

fragment
Exponent : ('e'|'E') ('+'|'-')? ('0'..'9')+ ;

fragment
FloatTypeSuffix : ('f'|'F'|'d'|'D'|'l'|'L') ;

fragment
EscapeSequence
	: '\\' ('a'|'b'|'f'|'n'|'r'|'t'|'v'|'\"'|'\''|'\\')
	| OctalEscape
	;

fragment
OctalEscape
	: '\\' ('0'..'3') ('0'..'7') ('0'..'7')
	| '\\' ('0'..'7') ('0'..'7')
	| '\\' ('0'..'7')
	;

fragment
UnicodeEscape
	: '\\' 'u' HexDigit HexDigit HexDigit HexDigit
	;

WS
	: (' '|'\r'|'\t'|'\u000C'|'\n')+ {SKIP();}
	;

COMMENT
	: '/*' .* '*/' {SKIP();}
	;

LINE_COMMENT
	: '//' ~('\n'|'\r')* '\r'? '\n' {SKIP();}
	;

LINE_COMMAND
	: '#' ~('\n'|'\r')* {$channel=HIDDEN;}
	;
